module Primer

open FStar.Mul

(* We always need this module to use Meta-F* *)
open FStar.Tactics

let idtac () : Tac unit = ()

let max_list (l1 l2 : list int) : Tac (list int) =
    if FStar.List.length l1 > FStar.List.length l2 then l1 else l2

let rec recursive_tac (n:nat) : Tac unit =
    if n > 0 then
      recursive_tac (n - 1)

let rec looping_tac () : Tac unit =
    // Metaprograms need not terminate!
    looping_tac ()

let _ = run_tactic (fun () -> dump "Hello world!")

#push-options "--tactic_trace_d 1"
let constr (a b : prop) : Lemma (a ==> b ==> b /\ a) =
  assert (a ==> b ==> b /\ a)
      by (let _ = implies_intro () in
          let _ = implies_intro () in
          split ();
          //hyp hb;
          //hyp ha;
          assumption ();
          assumption ();
          qed ())
#pop-options

(***** *****)

[@@expect_failure]
let pow2_bound_19 (x:nat{x <= 19}) : Lemma (pow2 x < 1000000) =
  assert (forall (x y : nat). x <= y ==> pow2 x <= pow2 y);
  assert (pow2 19 == 524288);
  assert (pow2 x < 1000000);
  ()

[@@expect_failure]
let pow2_bound_19' (x:nat{x <= 19}) : Lemma (pow2 x < 1000000) =
  FStar.Math.Lemmas.pow2_le_compat 19 x;
  assert (pow2 19 == 524288);
  assert (pow2 x < 1000000);
  ()

let pow2_bound_19'' (x:nat{x <= 19}) : Lemma (pow2 x < 1000000) =
  FStar.Math.Lemmas.pow2_le_compat 19 x;
  assert (pow2 19 == 524288) by (compute (); dump "After compute");
  assert (pow2 x < 1000000);
  ()

// let pow2_bound_19''' (x:nat{x <= 19}) : Lemma (pow2 x < 1000000) =
//   assert (pow2 19 == 524288) by compute ();
//   assert (pow2 x < 1000000) by (
//     let _ = using_lemma (`Math.Lemmas.pow2_le_compat) in
//     ()
//   );
//   ()

let _ =
  assert (True /\ True)
      by (match cur_formula () with
          | And _ _ -> dump "It's a conjunction!"
          | _ -> fail "Uh-oh!")

let _ =
  assert (False ==> True)
      by (let f = cur_formula () in
          match f with
          | Implies _ _ -> dump "It's an implication!"
          | _ -> fail "Uh-oh!")

let rec blowup () : Tac unit =
  match cur_formula () with
  | And _ _ -> split (); iseq [blowup; blowup]
  | Implies _ _ -> let _ = implies_intro () in blowup ()
  | _ -> ()

let test_blowup () =
  assert (True /\ (False ==> (True /\ False)))
      by (blowup (); dump "Final state")

#push-options "--tactic_trace_d 1"
let ex8 (p q r : prop) =
  assert ((p ==> q ==> r) ==> q ==> p ==> r)
      by (let i = implies_intro () in
          let pq = implies_intro () in
          let pp = implies_intro () in
          mapply i;
          mapply pp;
          mapply pq;
          qed ())
#pop-options

let ex9 (p : prop) =
  assert (p ==> (forall (_x:int). p))
      by (let bp = implies_intro () in
          let _bx = forall_intro () in
          hyp bp;
          qed ())

let ex10 (p : int -> prop) =
  assert ((forall x. p x) ==> (exists x. p x))
      by (let bf = implies_intro () in
          witness (`0);
          let bp0 = instantiate bf (`0) in
          hyp bp0;
          qed())

open FStar.Tactics.CanonCommSemiring

(* No smt at all required for this proof.
 * Also true of others! You can find them :-). *)
#push-options "--no_smt"
let arith_test (a b c d e : int) =
    assert
        ((a+b+c+d + e%2)*(b+c+d+a) ==
                a * a + b * b
              + c * c + d * d
              + e%2 * a + e%2 * b
              + e%2 * c + e%2 * d
              + a * b + a * b
              + a * c + a * c
              + a * d + a * d
              + b * c + b * c
              + b * d + b * d
              + c * d + c * d) by (
              dump "1";
                int_semiring ();
              dump "2";
                trefl();
                dump "After")
#pop-options

let lemma_poly_multiply (n p r h r0 r1 h0 h1 h2 s1 d0 d1 d2 hh : int)
  : Lemma
     (requires p > 0 /\ r1 >= 0 /\ n > 0 /\ 4 * (n * n) == p + 5 /\ r == r1 * n + r0 /\
               h == h2 * (n * n) + h1 * n + h0 /\ s1 == r1 + (r1 / 4) /\ r1 % 4 == 0 /\
               d0 == h0 * r0 + h1 * s1 /\ d1 == h0 * r1 + h1 * r0 + h2 * s1 /\
               d2 == h2 * r0 /\ hh == d2 * (n * n) + d1 * n + d0)
     (ensures (h * r) % p == hh % p)
  =
  let r1_4 = r1 / 4 in
  let h_r_expand = (h2 * (n * n) + h1 * n + h0) * ((r1_4 * 4) * n + r0) in
  let hh_expand = (h2 * r0) * (n * n) + (h0 * (r1_4 * 4) + h1 * r0 + h2 * (5 * r1_4)) * n
                  + (h0 * r0 + h1 * (5 * r1_4)) in
  let b = ((h2 * n + h1) * r1_4) in
  Math.Lemmas.modulo_addition_lemma hh_expand p b;
  assert (h_r_expand == hh_expand + b * (n * n * 4 + (-5)))
      by (canon_semiring int_cr)


(* Defines an identity function via metaprogramming *)
let fancy_id : x:int -> int = _ by (let x = intro () in
                               exact (binder_to_term x))

type enum = | A | B | C

let matchenum : enum -> int =
  _ by (let x = intro () in
        destruct (binder_to_term x);
        ignore (intro ()); exact (`1);
        ignore (intro ()); exact (`2);
        ignore (intro ()); exact (`3))

(* Calling `destruct` also allows to get the arguments of each constructor. *)
type either (a b : Type) = | Left of a | Right of b

let matcheither : either int int -> int =
  _ by (let x = intro () in
        destruct (binder_to_term x);
        let x = intro() in ignore (intro ()); exact (binder_to_term x);
        let y = intro() in ignore (intro ()); exact (binder_to_term y))

let _ = assert (matcheither (Left 10) == 10)
let _ = assert (matcheither (Right 20) == 20)

// examples/tactics/Printers.fst
// Metap.fst
